import re
from . import err


#: Regular expression for :meth:`Cursor.executemany`.
#: executemany only supports simple bulk insert.
#: You can use it to load large dataset.
RE_INSERT_VALUES = re.compile(
    r"\s*((?:INSERT|REPLACE)\b.+\bVALUES?\s*)"
    + r"(\(\s*(?:%s|%\(.+\)s)\s*(?:,\s*(?:%s|%\(.+\)s)\s*)*\))"
    + r"(\s*(?:ON DUPLICATE.*)?);?\s*\Z",
    re.IGNORECASE | re.DOTALL,
)


class Cursor:
    """
    This is the object you use to interact with the database.

    Do not create an instance of a Cursor yourself. Call
    connections.Connection.cursor().

    See `Cursor <https://www.python.org/dev/peps/pep-0249/#cursor-objects>`_ in
    the specification.
    """

    #: Max statement size which :meth:`executemany` generates.
    #:
    #: Max size of allowed statement is max_allowed_packet - packet_header_size.
    #: Default value of max_allowed_packet is 1048576.
    max_stmt_length = 1024000

    def __init__(self, connection):
        self.connection = connection
        self.description = None
        self.rownumber = 0
        self.rowcount = -1
        self.arraysize = 1
        self._executed = None
        self._result = None
        self._rows = None

    def close(self):
        """
        Closing a cursor just exhausts all remaining data.
        """
        conn = self.connection
        if conn is None:
            return
        try:
            while self.nextset():
                pass
        finally:
            self.connection = None

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        del exc_info
        self.close()

    def _get_db(self):
        if not self.connection:
            raise err.ProgrammingError("Cursor closed")
        return self.connection

    def _check_executed(self):
        if not self._executed:
            raise err.ProgrammingError("execute() first")

    def _conv_row(self, row):
        return row

    def setinputsizes(self, *args):
        """Does nothing, required by DB API."""

    def setoutputsizes(self, *args):
        """Does nothing, required by DB API."""

    def _nextset(self, unbuffered=False):
        """Get the next query set"""
        conn = self._get_db()
        current_result = self._result
        if current_result is None or current_result is not conn._result:
            return None
        if not current_result.has_next:
            return None
        self._result = None
        self._clear_result()
        conn.next_result(unbuffered=unbuffered)
        self._do_get_result()
        return True

    def nextset(self):
        return self._nextset(False)

    def _ensure_bytes(self, x, encoding=None):
        if isinstance(x, str):
            x = x.encode(encoding)
        elif isinstance(x, (tuple, list)):
            x = type(x)(self._ensure_bytes(v, encoding=encoding) for v in x)
        return x

    def _escape_args(self, args, conn):
        if isinstance(args, (tuple, list)):
            return tuple(conn.literal(arg) for arg in args)
        elif isinstance(args, dict):
            return {key: conn.literal(val) for (key, val) in args.items()}
        else:
            # If it's not a dictionary let's try escaping it anyways.
            # Worst case it will throw a Value error
            return conn.escape(args)

    def mogrify(self, query, args=None):
        """
        Returns the exact string that is sent to the database by calling the
        execute() method.

        This method follows the extension to the DB API 2.0 followed by Psycopg.
        """
        conn = self._get_db()

        if args is not None:
            query = query % self._escape_args(args, conn)

        return query

    def execute(self, query, args=None):
        """Execute a query

        :param str query: Query to execute.

        :param args: parameters used with query. (optional)
        :type args: tuple, list or dict

        :return: Number of affected rows
        :rtype: int

        If args is a list or tuple, %s can be used as a placeholder in the query.
        If args is a dict, %(name)s can be used as a placeholder in the query.
        """
        while self.nextset():
            pass

        query = self.mogrify(query, args)

        result = self._query(query)
        self._executed = query
        return result

    def executemany(self, query, args):
        # type: (str, list) -> int
        """Run several data against one query

        :param query: query to execute on server
        :param args:  Sequence of sequences or mappings.  It is used as parameter.
        :return: Number of rows affected, if any.

        This method improves performance on multiple-row INSERT and
        REPLACE. Otherwise it is equivalent to looping over args with
        execute().
        """
        if not args:
            return

        m = RE_INSERT_VALUES.match(query)
        if m:
            q_prefix = m.group(1) % ()
            q_values = m.group(2).rstrip()
            q_postfix = m.group(3) or ""
            assert q_values[0] == "(" and q_values[-1] == ")"
            return self._do_execute_many(
                q_prefix,
                q_values,
                q_postfix,
                args,
                self.max_stmt_length,
                self._get_db().encoding,
            )

        self.rowcount = sum(self.execute(query, arg) for arg in args)
        return self.rowcount

    def _do_execute_many(
        self, prefix, values, postfix, args, max_stmt_length, encoding
    ):
        conn = self._get_db()
        escape = self._escape_args
        if isinstance(prefix, str):
            prefix = prefix.encode(encoding)
        if isinstance(postfix, str):
            postfix = postfix.encode(encoding)
        sql = bytearray(prefix)
        args = iter(args)
        v = values % escape(next(args), conn)
        if isinstance(v, str):
            v = v.encode(encoding, "surrogateescape")
        sql += v
        rows = 0
        for arg in args:
            v = values % escape(arg, conn)
            if isinstance(v, str):
                v = v.encode(encoding, "surrogateescape")
            if len(sql) + len(v) + len(postfix) + 1 > max_stmt_length:
                rows += self.execute(sql + postfix)
                sql = bytearray(prefix)
            else:
                sql += b","
            sql += v
        rows += self.execute(sql + postfix)
        self.rowcount = rows
        return rows

    def callproc(self, procname, args=()):
        """Execute stored procedure procname with args

        procname -- string, name of procedure to execute on server

        args -- Sequence of parameters to use with procedure

        Returns the original args.

        Compatibility warning: PEP-249 specifies that any modified
        parameters must be returned. This is currently impossible
        as they are only available by storing them in a server
        variable and then retrieved by a query. Since stored
        procedures return zero or more result sets, there is no
        reliable way to get at OUT or INOUT parameters via callproc.
        The server variables are named @_procname_n, where procname
        is the parameter above and n is the position of the parameter
        (from zero). Once all result sets generated by the procedure
        have been fetched, you can issue a SELECT @_procname_0, ...
        query using .execute() to get any OUT or INOUT values.

        Compatibility warning: The act of calling a stored procedure
        itself creates an empty result set. This appears after any
        result sets generated by the procedure. This is non-standard
        behavior with respect to the DB-API. Be sure to use nextset()
        to advance through all result sets; otherwise you may get
        disconnected.
        """
        conn = self._get_db()
        if args:
            fmt = f"@_{procname}_%d=%s"
            self._query(
                "SET %s"
                % ",".join(
                    fmt % (index, conn.escape(arg)) for index, arg in enumerate(args)
                )
            )
            self.nextset()

        q = "CALL %s(%s)" % (
            procname,
            ",".join(["@_%s_%d" % (procname, i) for i in range(len(args))]),
        )
        self._query(q)
        self._executed = q
        return args

    def fetchone(self):
        """Fetch the next row"""
        self._check_executed()
        if self._rows is None or self.rownumber >= len(self._rows):
            return None
        result = self._rows[self.rownumber]
        self.rownumber += 1
        return result

    def fetchmany(self, size=None):
        """Fetch several rows"""
        self._check_executed()
        if self._rows is None:
            return ()
        end = self.rownumber + (size or self.arraysize)
        result = self._rows[self.rownumber : end]
        self.rownumber = min(end, len(self._rows))
        return result

    def fetchall(self):
        """Fetch all the rows"""
        self._check_executed()
        if self._rows is None:
            return ()
        if self.rownumber:
            result = self._rows[self.rownumber :]
        else:
            result = self._rows
        self.rownumber = len(self._rows)
        return result

    def scroll(self, value, mode="relative"):
        self._check_executed()
        if mode == "relative":
            r = self.rownumber + value
        elif mode == "absolute":
            r = value
        else:
            raise err.ProgrammingError("unknown scroll mode %s" % mode)

        if not (0 <= r < len(self._rows)):
            raise IndexError("out of range")
        self.rownumber = r

    def _query(self, q):
        conn = self._get_db()
        self._last_executed = q
        self._clear_result()
        conn.query(q)
        self._do_get_result()
        return self.rowcount

    def _clear_result(self):
        self.rownumber = 0
        self._result = None

        self.rowcount = 0
        self.description = None
        self.lastrowid = None
        self._rows = None

    def _do_get_result(self):
        conn = self._get_db()

        self._result = result = conn._result

        self.rowcount = result.affected_rows
        self.description = result.description
        self.lastrowid = result.insert_id
        self._rows = result.rows

    def __iter__(self):
        return iter(self.fetchone, None)

    Warning = err.Warning
    Error = err.Error
    InterfaceError = err.InterfaceError
    DatabaseError = err.DatabaseError
    DataError = err.DataError
    OperationalError = err.OperationalError
    IntegrityError = err.IntegrityError
    InternalError = err.InternalError
    ProgrammingError = err.ProgrammingError
    NotSupportedError = err.NotSupportedError


class DictCursorMixin:
    # You can override this to use OrderedDict or other dict-like types.
    dict_type = dict

    def _do_get_result(self):
        super(DictCursorMixin, self)._do_get_result()
        fields = []
        if self.description:
            for f in self._result.fields:
                name = f.name
                if name in fields:
                    name = f.table_name + "." + name
                fields.append(name)
            self._fields = fields

        if fields and self._rows:
            self._rows = [self._conv_row(r) for r in self._rows]

    def _conv_row(self, row):
        if row is None:
            return None
        return self.dict_type(zip(self._fields, row))


class DictCursor(DictCursorMixin, Cursor):
    """A cursor which returns results as a dictionary"""


class SSCursor(Cursor):
    """
    Unbuffered Cursor, mainly useful for queries that return a lot of data,
    or for connections to remote servers over a slow network.

    Instead of copying every row of data into a buffer, this will fetch
    rows as needed. The upside of this is the client uses much less memory,
    and rows are returned much faster when traveling over a slow network
    or if the result set is very big.

    There are limitations, though. The MySQL protocol doesn't support
    returning the total number of rows, so the only way to tell how many rows
    there are is to iterate over every row returned. Also, it currently isn't
    possible to scroll backwards, as only the current row is held in memory.
    """

    def _conv_row(self, row):
        return row

    def close(self):
        conn = self.connection
        if conn is None:
            return

        if self._result is not None and self._result is conn._result:
            self._result._finish_unbuffered_query()

        try:
            while self.nextset():
                pass
        finally:
            self.connection = None

    __del__ = close

    def _query(self, q):
        conn = self._get_db()
        self._last_executed = q
        self._clear_result()
        conn.query(q, unbuffered=True)
        self._do_get_result()
        return self.rowcount

    def nextset(self):
        return self._nextset(unbuffered=True)

    def read_next(self):
        """Read next row"""
        return self._conv_row(self._result._read_rowdata_packet_unbuffered())

    def fetchone(self):
        """Fetch next row"""
        self._check_executed()
        row = self.read_next()
        if row is None:
            return None
        self.rownumber += 1
        return row

    def fetchall(self):
        """
        Fetch all, as per MySQLdb. Pretty useless for large queries, as
        it is buffered. See fetchall_unbuffered(), if you want an unbuffered
        generator version of this method.
        """
        return list(self.fetchall_unbuffered())

    def fetchall_unbuffered(self):
        """
        Fetch all, implemented as a generator, which isn't to standard,
        however, it doesn't make sense to return everything in a list, as that
        would use ridiculous memory for large result sets.
        """
        return iter(self.fetchone, None)

    def __iter__(self):
        return self.fetchall_unbuffered()

    def fetchmany(self, size=None):
        """Fetch many"""
        self._check_executed()
        if size is None:
            size = self.arraysize

        rows = []
        for i in range(size):
            row = self.read_next()
            if row is None:
                break
            rows.append(row)
            self.rownumber += 1
        return rows

    def scroll(self, value, mode="relative"):
        self._check_executed()

        if mode == "relative":
            if value < 0:
                raise err.NotSupportedError(
                    "Backwards scrolling not supported by this cursor"
                )

            for _ in range(value):
                self.read_next()
            self.rownumber += value
        elif mode == "absolute":
            if value < self.rownumber:
                raise err.NotSupportedError(
                    "Backwards scrolling not supported by this cursor"
                )

            end = value - self.rownumber
            for _ in range(end):
                self.read_next()
            self.rownumber = value
        else:
            raise err.ProgrammingError("unknown scroll mode %s" % mode)


class SSDictCursor(DictCursorMixin, SSCursor):
    """An unbuffered cursor, which returns results as a dictionary"""



MBLENGTH = {8: 1, 33: 3, 88: 2, 91: 2}


class Charset:
    def __init__(self, id, name, collation, is_default):
        self.id, self.name, self.collation = id, name, collation
        self.is_default = is_default == "Yes"

    def __repr__(self):
        return "Charset(id=%s, name=%r, collation=%r)" % (
            self.id,
            self.name,
            self.collation,
        )

    @property
    def encoding(self):
        name = self.name
        if name in ("utf8mb4", "utf8mb3"):
            return "utf8"
        if name == "latin1":
            return "cp1252"
        if name == "koi8r":
            return "koi8_r"
        if name == "koi8u":
            return "koi8_u"
        return name

    @property
    def is_binary(self):
        return self.id == 63


class Charsets:
    def __init__(self):
        self._by_id = {}
        self._by_name = {}

    def add(self, c):
        self._by_id[c.id] = c
        if c.is_default:
            self._by_name[c.name] = c

    def by_id(self, id):
        return self._by_id[id]

    def by_name(self, name):
        return self._by_name.get(name.lower())


_charsets = Charsets()
"""
Generated with:

mysql -N -s -e "select id, character_set_name, collation_name, is_default
from information_schema.collations order by id;" | python -c "import sys
for l in sys.stdin.readlines():
        id, name, collation, is_default  = l.split(chr(9))
        print '_charsets.add(Charset(%s, \'%s\', \'%s\', \'%s\'))' \
                % (id, name, collation, is_default.strip())
"

"""
_charsets.add(Charset(1, "big5", "big5_chinese_ci", "Yes"))
_charsets.add(Charset(2, "latin2", "latin2_czech_cs", ""))
_charsets.add(Charset(3, "dec8", "dec8_swedish_ci", "Yes"))
_charsets.add(Charset(4, "cp850", "cp850_general_ci", "Yes"))
_charsets.add(Charset(5, "latin1", "latin1_german1_ci", ""))
_charsets.add(Charset(6, "hp8", "hp8_english_ci", "Yes"))
_charsets.add(Charset(7, "koi8r", "koi8r_general_ci", "Yes"))
_charsets.add(Charset(8, "latin1", "latin1_swedish_ci", "Yes"))
_charsets.add(Charset(9, "latin2", "latin2_general_ci", "Yes"))
_charsets.add(Charset(10, "swe7", "swe7_swedish_ci", "Yes"))
_charsets.add(Charset(11, "ascii", "ascii_general_ci", "Yes"))
_charsets.add(Charset(12, "ujis", "ujis_japanese_ci", "Yes"))
_charsets.add(Charset(13, "sjis", "sjis_japanese_ci", "Yes"))
_charsets.add(Charset(14, "cp1251", "cp1251_bulgarian_ci", ""))
_charsets.add(Charset(15, "latin1", "latin1_danish_ci", ""))
_charsets.add(Charset(16, "hebrew", "hebrew_general_ci", "Yes"))
_charsets.add(Charset(18, "tis620", "tis620_thai_ci", "Yes"))
_charsets.add(Charset(19, "euckr", "euckr_korean_ci", "Yes"))
_charsets.add(Charset(20, "latin7", "latin7_estonian_cs", ""))
_charsets.add(Charset(21, "latin2", "latin2_hungarian_ci", ""))
_charsets.add(Charset(22, "koi8u", "koi8u_general_ci", "Yes"))
_charsets.add(Charset(23, "cp1251", "cp1251_ukrainian_ci", ""))
_charsets.add(Charset(24, "gb2312", "gb2312_chinese_ci", "Yes"))
_charsets.add(Charset(25, "greek", "greek_general_ci", "Yes"))
_charsets.add(Charset(26, "cp1250", "cp1250_general_ci", "Yes"))
_charsets.add(Charset(27, "latin2", "latin2_croatian_ci", ""))
_charsets.add(Charset(28, "gbk", "gbk_chinese_ci", "Yes"))
_charsets.add(Charset(29, "cp1257", "cp1257_lithuanian_ci", ""))
_charsets.add(Charset(30, "latin5", "latin5_turkish_ci", "Yes"))
_charsets.add(Charset(31, "latin1", "latin1_german2_ci", ""))
_charsets.add(Charset(32, "armscii8", "armscii8_general_ci", "Yes"))
_charsets.add(Charset(33, "utf8", "utf8_general_ci", "Yes"))
_charsets.add(Charset(34, "cp1250", "cp1250_czech_cs", ""))
_charsets.add(Charset(36, "cp866", "cp866_general_ci", "Yes"))
_charsets.add(Charset(37, "keybcs2", "keybcs2_general_ci", "Yes"))
_charsets.add(Charset(38, "macce", "macce_general_ci", "Yes"))
_charsets.add(Charset(39, "macroman", "macroman_general_ci", "Yes"))
_charsets.add(Charset(40, "cp852", "cp852_general_ci", "Yes"))
_charsets.add(Charset(41, "latin7", "latin7_general_ci", "Yes"))
_charsets.add(Charset(42, "latin7", "latin7_general_cs", ""))
_charsets.add(Charset(43, "macce", "macce_bin", ""))
_charsets.add(Charset(44, "cp1250", "cp1250_croatian_ci", ""))
_charsets.add(Charset(45, "utf8mb4", "utf8mb4_general_ci", "Yes"))
_charsets.add(Charset(46, "utf8mb4", "utf8mb4_bin", ""))
_charsets.add(Charset(47, "latin1", "latin1_bin", ""))
_charsets.add(Charset(48, "latin1", "latin1_general_ci", ""))
_charsets.add(Charset(49, "latin1", "latin1_general_cs", ""))
_charsets.add(Charset(50, "cp1251", "cp1251_bin", ""))
_charsets.add(Charset(51, "cp1251", "cp1251_general_ci", "Yes"))
_charsets.add(Charset(52, "cp1251", "cp1251_general_cs", ""))
_charsets.add(Charset(53, "macroman", "macroman_bin", ""))
_charsets.add(Charset(57, "cp1256", "cp1256_general_ci", "Yes"))
_charsets.add(Charset(58, "cp1257", "cp1257_bin", ""))
_charsets.add(Charset(59, "cp1257", "cp1257_general_ci", "Yes"))
_charsets.add(Charset(63, "binary", "binary", "Yes"))
_charsets.add(Charset(64, "armscii8", "armscii8_bin", ""))
_charsets.add(Charset(65, "ascii", "ascii_bin", ""))
_charsets.add(Charset(66, "cp1250", "cp1250_bin", ""))
_charsets.add(Charset(67, "cp1256", "cp1256_bin", ""))
_charsets.add(Charset(68, "cp866", "cp866_bin", ""))
_charsets.add(Charset(69, "dec8", "dec8_bin", ""))
_charsets.add(Charset(70, "greek", "greek_bin", ""))
_charsets.add(Charset(71, "hebrew", "hebrew_bin", ""))
_charsets.add(Charset(72, "hp8", "hp8_bin", ""))
_charsets.add(Charset(73, "keybcs2", "keybcs2_bin", ""))
_charsets.add(Charset(74, "koi8r", "koi8r_bin", ""))
_charsets.add(Charset(75, "koi8u", "koi8u_bin", ""))
_charsets.add(Charset(76, "utf8", "utf8_tolower_ci", ""))
_charsets.add(Charset(77, "latin2", "latin2_bin", ""))
_charsets.add(Charset(78, "latin5", "latin5_bin", ""))
_charsets.add(Charset(79, "latin7", "latin7_bin", ""))
_charsets.add(Charset(80, "cp850", "cp850_bin", ""))
_charsets.add(Charset(81, "cp852", "cp852_bin", ""))
_charsets.add(Charset(82, "swe7", "swe7_bin", ""))
_charsets.add(Charset(83, "utf8", "utf8_bin", ""))
_charsets.add(Charset(84, "big5", "big5_bin", ""))
_charsets.add(Charset(85, "euckr", "euckr_bin", ""))
_charsets.add(Charset(86, "gb2312", "gb2312_bin", ""))
_charsets.add(Charset(87, "gbk", "gbk_bin", ""))
_charsets.add(Charset(88, "sjis", "sjis_bin", ""))
_charsets.add(Charset(89, "tis620", "tis620_bin", ""))
_charsets.add(Charset(91, "ujis", "ujis_bin", ""))
_charsets.add(Charset(92, "geostd8", "geostd8_general_ci", "Yes"))
_charsets.add(Charset(93, "geostd8", "geostd8_bin", ""))
_charsets.add(Charset(94, "latin1", "latin1_spanish_ci", ""))
_charsets.add(Charset(95, "cp932", "cp932_japanese_ci", "Yes"))
_charsets.add(Charset(96, "cp932", "cp932_bin", ""))
_charsets.add(Charset(97, "eucjpms", "eucjpms_japanese_ci", "Yes"))
_charsets.add(Charset(98, "eucjpms", "eucjpms_bin", ""))
_charsets.add(Charset(99, "cp1250", "cp1250_polish_ci", ""))
_charsets.add(Charset(192, "utf8", "utf8_unicode_ci", ""))
_charsets.add(Charset(193, "utf8", "utf8_icelandic_ci", ""))
_charsets.add(Charset(194, "utf8", "utf8_latvian_ci", ""))
_charsets.add(Charset(195, "utf8", "utf8_romanian_ci", ""))
_charsets.add(Charset(196, "utf8", "utf8_slovenian_ci", ""))
_charsets.add(Charset(197, "utf8", "utf8_polish_ci", ""))
_charsets.add(Charset(198, "utf8", "utf8_estonian_ci", ""))
_charsets.add(Charset(199, "utf8", "utf8_spanish_ci", ""))
_charsets.add(Charset(200, "utf8", "utf8_swedish_ci", ""))
_charsets.add(Charset(201, "utf8", "utf8_turkish_ci", ""))
_charsets.add(Charset(202, "utf8", "utf8_czech_ci", ""))
_charsets.add(Charset(203, "utf8", "utf8_danish_ci", ""))
_charsets.add(Charset(204, "utf8", "utf8_lithuanian_ci", ""))
_charsets.add(Charset(205, "utf8", "utf8_slovak_ci", ""))
_charsets.add(Charset(206, "utf8", "utf8_spanish2_ci", ""))
_charsets.add(Charset(207, "utf8", "utf8_roman_ci", ""))
_charsets.add(Charset(208, "utf8", "utf8_persian_ci", ""))
_charsets.add(Charset(209, "utf8", "utf8_esperanto_ci", ""))
_charsets.add(Charset(210, "utf8", "utf8_hungarian_ci", ""))
_charsets.add(Charset(211, "utf8", "utf8_sinhala_ci", ""))
_charsets.add(Charset(212, "utf8", "utf8_german2_ci", ""))
_charsets.add(Charset(213, "utf8", "utf8_croatian_ci", ""))
_charsets.add(Charset(214, "utf8", "utf8_unicode_520_ci", ""))
_charsets.add(Charset(215, "utf8", "utf8_vietnamese_ci", ""))
_charsets.add(Charset(223, "utf8", "utf8_general_mysql500_ci", ""))
_charsets.add(Charset(224, "utf8mb4", "utf8mb4_unicode_ci", ""))
_charsets.add(Charset(225, "utf8mb4", "utf8mb4_icelandic_ci", ""))
_charsets.add(Charset(226, "utf8mb4", "utf8mb4_latvian_ci", ""))
_charsets.add(Charset(227, "utf8mb4", "utf8mb4_romanian_ci", ""))
_charsets.add(Charset(228, "utf8mb4", "utf8mb4_slovenian_ci", ""))
_charsets.add(Charset(229, "utf8mb4", "utf8mb4_polish_ci", ""))
_charsets.add(Charset(230, "utf8mb4", "utf8mb4_estonian_ci", ""))
_charsets.add(Charset(231, "utf8mb4", "utf8mb4_spanish_ci", ""))
_charsets.add(Charset(232, "utf8mb4", "utf8mb4_swedish_ci", ""))
_charsets.add(Charset(233, "utf8mb4", "utf8mb4_turkish_ci", ""))
_charsets.add(Charset(234, "utf8mb4", "utf8mb4_czech_ci", ""))
_charsets.add(Charset(235, "utf8mb4", "utf8mb4_danish_ci", ""))
_charsets.add(Charset(236, "utf8mb4", "utf8mb4_lithuanian_ci", ""))
_charsets.add(Charset(237, "utf8mb4", "utf8mb4_slovak_ci", ""))
_charsets.add(Charset(238, "utf8mb4", "utf8mb4_spanish2_ci", ""))
_charsets.add(Charset(239, "utf8mb4", "utf8mb4_roman_ci", ""))
_charsets.add(Charset(240, "utf8mb4", "utf8mb4_persian_ci", ""))
_charsets.add(Charset(241, "utf8mb4", "utf8mb4_esperanto_ci", ""))
_charsets.add(Charset(242, "utf8mb4", "utf8mb4_hungarian_ci", ""))
_charsets.add(Charset(243, "utf8mb4", "utf8mb4_sinhala_ci", ""))
_charsets.add(Charset(244, "utf8mb4", "utf8mb4_german2_ci", ""))
_charsets.add(Charset(245, "utf8mb4", "utf8mb4_croatian_ci", ""))
_charsets.add(Charset(246, "utf8mb4", "utf8mb4_unicode_520_ci", ""))
_charsets.add(Charset(247, "utf8mb4", "utf8mb4_vietnamese_ci", ""))
_charsets.add(Charset(248, "gb18030", "gb18030_chinese_ci", "Yes"))
_charsets.add(Charset(249, "gb18030", "gb18030_bin", ""))
_charsets.add(Charset(250, "gb18030", "gb18030_unicode_520_ci", ""))
_charsets.add(Charset(255, "utf8mb4", "utf8mb4_0900_ai_ci", ""))

charset_by_name = _charsets.by_name
charset_by_id = _charsets.by_id



# flake8: noqa
# errmsg.h
CR_ERROR_FIRST = 2000
CR_UNKNOWN_ERROR = 2000
CR_SOCKET_CREATE_ERROR = 2001
CR_CONNECTION_ERROR = 2002
CR_CONN_HOST_ERROR = 2003
CR_IPSOCK_ERROR = 2004
CR_UNKNOWN_HOST = 2005
CR_SERVER_GONE_ERROR = 2006
CR_VERSION_ERROR = 2007
CR_OUT_OF_MEMORY = 2008
CR_WRONG_HOST_INFO = 2009
CR_LOCALHOST_CONNECTION = 2010
CR_TCP_CONNECTION = 2011
CR_SERVER_HANDSHAKE_ERR = 2012
CR_SERVER_LOST = 2013
CR_COMMANDS_OUT_OF_SYNC = 2014
CR_NAMEDPIPE_CONNECTION = 2015
CR_NAMEDPIPEWAIT_ERROR = 2016
CR_NAMEDPIPEOPEN_ERROR = 2017
CR_NAMEDPIPESETSTATE_ERROR = 2018
CR_CANT_READ_CHARSET = 2019
CR_NET_PACKET_TOO_LARGE = 2020
CR_EMBEDDED_CONNECTION = 2021
CR_PROBE_SLAVE_STATUS = 2022
CR_PROBE_SLAVE_HOSTS = 2023
CR_PROBE_SLAVE_CONNECT = 2024
CR_PROBE_MASTER_CONNECT = 2025
CR_SSL_CONNECTION_ERROR = 2026
CR_MALFORMED_PACKET = 2027
CR_WRONG_LICENSE = 2028

CR_NULL_POINTER = 2029
CR_NO_PREPARE_STMT = 2030
CR_PARAMS_NOT_BOUND = 2031
CR_DATA_TRUNCATED = 2032
CR_NO_PARAMETERS_EXISTS = 2033
CR_INVALID_PARAMETER_NO = 2034
CR_INVALID_BUFFER_USE = 2035
CR_UNSUPPORTED_PARAM_TYPE = 2036

CR_SHARED_MEMORY_CONNECTION = 2037
CR_SHARED_MEMORY_CONNECT_REQUEST_ERROR = 2038
CR_SHARED_MEMORY_CONNECT_ANSWER_ERROR = 2039
CR_SHARED_MEMORY_CONNECT_FILE_MAP_ERROR = 2040
CR_SHARED_MEMORY_CONNECT_MAP_ERROR = 2041
CR_SHARED_MEMORY_FILE_MAP_ERROR = 2042
CR_SHARED_MEMORY_MAP_ERROR = 2043
CR_SHARED_MEMORY_EVENT_ERROR = 2044
CR_SHARED_MEMORY_CONNECT_ABANDONED_ERROR = 2045
CR_SHARED_MEMORY_CONNECT_SET_ERROR = 2046
CR_CONN_UNKNOW_PROTOCOL = 2047
CR_INVALID_CONN_HANDLE = 2048
CR_SECURE_AUTH = 2049
CR_FETCH_CANCELED = 2050
CR_NO_DATA = 2051
CR_NO_STMT_METADATA = 2052
CR_NO_RESULT_SET = 2053
CR_NOT_IMPLEMENTED = 2054
CR_SERVER_LOST_EXTENDED = 2055
CR_STMT_CLOSED = 2056
CR_NEW_STMT_METADATA = 2057
CR_ALREADY_CONNECTED = 2058
CR_AUTH_PLUGIN_CANNOT_LOAD = 2059
CR_DUPLICATE_CONNECTION_ATTR = 2060
CR_AUTH_PLUGIN_ERR = 2061
CR_ERROR_LAST = 2061
